@@ -34,7 +34,6 @@
 """ Script to run external tests, eg from
 https://github.com/json-patch/json-patch-tests """
 
-from functools import partial
 import doctest
 import unittest
 import jsonpatch
@@ -60,12 +59,15 @@ class TestCaseTemplate(unittest.TestCase):
                 )
 
         else:
-            res = jsonpatch.apply_patch(test['doc'], test['patch'])
+            try:
+                res = jsonpatch.apply_patch(test['doc'], test['patch'])
+            except jsonpatch.JsonPatchException as jpe:
+                raise Exception(test.get('comment', '')) from jpe
 
             # if there is no 'expected' we only verify that applying the patch
             # does not raies an exception
             if 'expected' in test:
-                self.assertEquals(res, test['expected'])
+                self.assertEquals(res, test['expected'], test.get('comment', ''))
 
 
 def make_test_case(tests):
@@ -112,24 +114,25 @@ runner = unittest.TextTestRunner(verbosity=1)
 
 try:
     import coverage
+    cov = coverage.Coverage()
 except ImportError:
-    coverage = None
+    cov = None
 
-if coverage is not None:
-    coverage.erase()
-    coverage.start()
+if cov is not None:
+    cov.erase()
+    cov.start()
 
 result = runner.run(suite)
 
 if not result.wasSuccessful():
     sys.exit(1)
 
-if coverage is not None:
-    coverage.stop()
-    coverage.report(coverage_modules)
-    coverage.erase()
+if cov is not None:
+    cov.stop()
+    cov.report(coverage_modules)
+    cov.erase()
 
-if coverage is None:
+if cov is None:
     sys.stderr.write("""
 No coverage reporting done (Python module "coverage" is missing)
 Please install the python-coverage package to get coverage reporting.
